class NestError(Exception):
    """Base class for all workflow exceptions"""


class WorkflowExistsError(NestError):
    """Thrown if the given workflow is already installed"""


class WorkflowNotFoundError(NestError):
    """Thrown if the given workflow cannot be found"""


class InvalidWorkflowRepositoryError(NestError):
    """Thrown if the given repository appears to have an invalid format."""


class SnkConfigError(Exception):
    """Base class for all SNK config exceptions"""

class InvalidSnkConfigError(SnkConfigError, ValueError):
    """Thrown if the given SNK config appears to have an invalid format."""

class MissingSnkConfigError(SnkConfigError, FileNotFoundError):
    """Thrown if the given SNK config file cannot be found."""


from pathlib import Path
from snk.cli.validate import validate_config
from snk.cli.config import SnkConfig
import pytest

@pytest.mark.parametrize("config,annotations,expected", [
    (
        {"example": "1"},
        {"example": {"type": "int"}},
        {"example": 1}
    ),
    (
        {"example": 1},
        {"example": {"type": "str"}},
        {"example": "1"}
    ),
    (
        {"example": "1"},
        {"example": {"type": "float"}},
        {"example": 1.0}
    ),
    (
        {"example": "1"},
        {"example": {"type": "bool"}},
        {"example": True}
    ),
    # None
    (
        {"example": None},
        {"example": {"type": "int"}},
        {"example": None}
    ),
    (
        {"example": None},
        {"example": {"type": "str"}},
        {"example": None}
    ),
    (
        {"example": None},
        {"example": {"type": "float"}},
        {"example": None}
    ),
    (
        {"example": None},
        {"example": {"type": "bool"}},
        {"example": None}
    ),
    # path type
    (
        {"example": "file"},
        {"example": {"type": "path"}},
        {"example": Path("file")}
    ),
    # list type
    (
        {"example": ["1", "2"]},
        {"example": {"type": "list[int]"}},
        {"example": [1, 2]}
    ),
    (
        {"example": ["1", "2"]},
        {"example": {"type": "list[str]"}},
        {"example": ["1", "2"]}
    ),
    (
        {"example": ["1", "2"]},
        {"example": {"type": "list[path]"}},
        {"example": [Path("1"), Path("2")]}
    ),
    # nested dictionary
    (
        {"example": {"nested": "1"}},
        {"example": {"nested": {"type": "int"}}},
        {"example": {"nested": 1}}
    ),
    # exception
    (
        {"example": "1"},
        {"example": {"type": "unknown"}},
        ValueError
    ),
    (
        {"example": "s"},
        {"example": {"type": "int"}},
        Exception
    ),
    (
        {"example": "1"},
        {"example": {"type": "list[int]"}},
        Exception
    ),
    (
        {"example": {"nested": "1"}},
        {"example": {"nested": {"type": "list[int]"}}},
        Exception
    ),
    ])
def test_validate_config(tmp_path, config, annotations, expected):
    """
    Tests for validate_config().
    """
    snk_config_path = tmp_path / "snk.yaml"
    SnkConfig(cli=annotations).to_yaml(snk_config_path)
    if isinstance(expected, type):
        with pytest.raises(expected):
            validate_config(config, snk_config_path)
    else:
        validate_config(config, snk_config_path)
        assert config == expected


from pathlib import Path
from typing import Any, Dict, Union, List
from snk.cli.config import SnkConfig
from snk.cli.options.utils import types
import inspect

class ValidationError(Exception):
    """Base class for all validation exceptions"""


def validate_config(config: Dict[str, Any], snk_config_path: Path) -> None:
    """
    Validates the config against the snk config.
    Will convert values to the correct type if possible.
    Args:
        config (dict): The config to validate.
        snk_config_path (Path): The path to the snk config.
    """
    snk_config_path = Path(snk_config_path)
    # if relative path, make absolute to 
    if not snk_config_path.is_absolute():
        frame = inspect.currentframe().f_back
        workflow = frame.f_globals.get("workflow")
        snk_config_path = Path(workflow.current_basedir) / snk_config_path

    snk_config = SnkConfig.from_path(snk_config_path)
    validate_and_transform_in_place(config, snk_config.cli)

ValidationDict = Dict[str, Union["ValidationDict", Dict[str, str]]]

def validate_and_transform_in_place(config: Dict[str, Any], validation: ValidationDict, replace_none: bool = True) -> None:
    """
    Validates the config against the snk config.
    Will convert values to the correct type if possible.
    Args:
        config (dict): The config to validate.
        validation (dict): The validation dict.
        replace_none (bool): If True, replace 'None' with None.
    """
    for key, value in list(config.items()):
        if key not in validation:
            continue  # Optionally handle unexpected keys
        if value == 'None' and replace_none:
            config[key] = None
            continue
        if value is None:
            continue
        val_info = validation[key]
        if isinstance(val_info, dict) and 'type' in val_info:
            # Direct type validation
            val_type = types.get(val_info["type"], None)
            if val_type is None:
                raise ValueError(f"Unknown type '{val_info['type']}'")
            try:
                if getattr(val_type, "__origin__", None) == list:
                    val_type = val_type.__args__[0]
                    if not isinstance(value, list):
                        raise ValueError(f"Expected a list for key '{key}'")
                    config[key] = [val_type(v) for v in value]
                else:
                    config[key] = val_type(value)
            except (ValueError, TypeError) as e:
                raise ValueError(f"Type conversion error for key '{key}': {e}")
        elif isinstance(value, dict):
            # Nested dictionary validation
            validate_and_transform_in_place(value, val_info)


from typing import List, Any
from ..config.config import SnkConfig
from ..utils import get_default_type, flatten
from .option import Option
from pathlib import Path

types = {
    "int": int,
    "integer": int,
    "float": float,
    "str": str,
    "string": str,
    "path": Path,
    "bool": bool,
    "boolean": bool,
    "list": List[str],
    "list[str]": List[str],
    "list[path]": List[Path],
    "list[int]": List[int],
}

def get_keys_from_annotation(annotations):
    return {":".join(annotation.split(":")[:-1]) for annotation in annotations}

def create_option_from_annotation(
    annotation_key: str,
    annotation_values: dict,
    default_values: dict,
    from_annotation: bool = False,
) -> Option:
    """
    Create an Option object from a given annotation.
    Args:
      annotation_key: The key in the annotations.
      annotation_values: The dictionary of annotation values.
      default_values: default value from config.
    Returns:
      An Option object.
    """    
    config_default = default_values.get(annotation_key, None)

    default = annotation_values.get(f"{annotation_key}:default", config_default)
    updated = False
    if config_default is None or default != config_default:
        updated = True
    type = annotation_values.get(f"{annotation_key}:type", get_default_type(default))
    assert (
        type is not None
    ), f"Type for {annotation_key} should be one of {', '.join(types.keys())}."
    annotation_type = types.get(
        type.lower(), List[str] if "list" in type.lower() else str
    )
    name = annotation_values.get(
        f"{annotation_key}:name", annotation_key.replace(":", "_")
    ).replace("-", "_")
    short = annotation_values.get(f"{annotation_key}:short", None)
    hidden = annotation_values.get(f"{annotation_key}:hidden", False)
    return Option(
        name=name,
        original_key=annotation_key,
        default=annotation_values.get(f"{annotation_key}:default", default),
        updated=updated,
        help=annotation_values.get(f"{annotation_key}:help", ""),
        type=annotation_type,
        required=annotation_values.get(f"{annotation_key}:required", False),
        short=short,
        flag=f"--{name.replace('_', '-')}",
        short_flag=f"-{short}" if short else None,
        hidden=hidden,
        from_annotation=from_annotation,
    )


def build_dynamic_cli_options(
    snakemake_config: dict, snk_config: SnkConfig
) -> List[dict]:
    """
    Builds a list of options from a snakemake config and a snk config.
    Args:
      snakemake_config (dict): A snakemake config.
      snk_config (SnkConfig): A snk config.
    Returns:
      List[dict]: A list of options.
    """
    flat_config = flatten(snakemake_config)
    flat_annotations = flatten(snk_config.cli)
    annotation_keys = get_keys_from_annotation(flat_annotations)
    options = {}

    # For every parameter in the config, create an option from the corresponding annotation
    for parameter in flat_config:
        if parameter not in annotation_keys and snk_config.skip_missing:
            continue
        options[parameter] = create_option_from_annotation(
            parameter,
            flat_annotations,
            default_values=flat_config,
        )

    # For every annotation not in config, create an option with default values    
    for key in annotation_keys:
        if key not in options:
            # in annotation but not in config
            options[key] = create_option_from_annotation(
                key,
                flat_annotations,
                default_values={},
                from_annotation=True,
            )
    return list(options.values())
